from abc import ABC, abstractmethod
from typing import Any, List
from urllib.parse import quote

from harambe_core.types import URL, Context, Options, Cookie, LocalStorage
from .base import OutputObserver
from .types import DownloadMeta


class DataTracker(ABC):
    @abstractmethod
    def save_data(self, new_data: List[dict[str, Any]]) -> None:
        """Append data for a domain and stage to a JSON file"""
        raise NotImplementedError()

    @abstractmethod
    def load_data(self, url: str, stage: str) -> List[dict[str, Any]]:
        """Load data for a domain and stage from a JSON file"""
        raise NotImplementedError()

    @abstractmethod
    def visit(self, url: str) -> None:
        """Mark the url as visited and save to the domain file"""
        raise NotImplementedError()

    @abstractmethod
    def has_been_visited(self, url: str) -> bool:
        """Check if the url has been visited"""
        raise NotImplementedError()


class LocalStorageObserver(OutputObserver):
    def __init__(self, tracker: DataTracker):
        self._tracker = tracker

    async def on_save_data(self, data: dict[str, Any]) -> None:
        self._tracker.save_data(data)

    async def on_queue_url(self, url: URL, context: Context, options: Options) -> None:
        self._tracker.save_data({"url": url, "context": context, "options": options})

    async def on_download(
        self, download_url: str, filename: str, content: bytes, path: str
    ) -> DownloadMeta:
        data: DownloadMeta = {
            "url": f"{download_url}/{quote(filename)}",
            "filename": filename,
            "path": path,
        }
        self._tracker.save_data(data)  # type: ignore
        return data

    async def on_paginate(self, next_url: str) -> None:
        pass

    async def on_save_cookies(self, cookies: List[Cookie]) -> None:
        self._tracker.save_data({"cookies": cookies})

    async def on_save_local_storage(self, local_storage: List[LocalStorage]) -> None:
        self._tracker.save_data({"local_storage": local_storage})

    async def on_check_and_solve_captchas(self, page: "Page") -> None:
        pass



from typing import Literal, TypedDict

ObservationTrigger = Literal[
    "on_save_data",
    "on_queue_url",
    "on_download",
    "on_paginate",
    "on_save_cookies",
    "on_save_local_storage",
    "on_check_and_solve_captchas",
]


class DownloadMeta(TypedDict):
    url: str
    filename: str
    path: str


class HTMLMetadata(DownloadMeta):
    html: str
    text: str



from typing import Any, Callable

from pydantic import BeforeValidator
from typing_extensions import Annotated, Literal


class ParserTypeEnum:
    def __new__(cls, *variants: str, required: bool = True) -> Any:
        validator = BeforeValidator(cls.validate_type(*variants))
        base = Literal[*variants] if required else Literal[*variants] | None

        return Annotated[base, validator]

    @staticmethod
    def validate_type(*variants: str) -> Callable[[Any], str]:
        variant_map = {v.strip().lower(): v for v in variants}

        def _validate_type(value: Any) -> Any:
            if not isinstance(value, str):
                return value

            if (value := value.strip().lower()) not in variant_map:
                return value

            return variant_map[value]

        return _validate_type



import pytest

from harambe_core.normalize_url import normalize_url, sanitize_scheme


@pytest.mark.parametrize(
    "base_path, url, expected",
    [
        (
            "https://example.com/subdir/page.html",
            "/x/X/picture.png",
            "https://example.com/x/X/picture.png",
        ),
        (
            "https://example.com/subdir/page.html",
            "./../x/X/picture.png",
            "https://example.com/x/X/picture.png",
        ),
        (
            "https://example.com/",
            "https://example.com/x/X/picture.png",
            "https://example.com/x/X/picture.png",
        ),
        ("https://example.com/subdir/", "", "https://example.com/subdir/"),
        (
            "https://example.com/",
            "/x/X/picture.png?query=123#fragment",
            "https://example.com/x/X/picture.png?query=123#fragment",
        ),
        (
            "https://example.com/subdir/",
            "./../../x/X/picture.png",
            "https://example.com/x/X/picture.png",
        ),
        (
            "https://example.com/subdir/",
            "./img/picture.png",
            "https://example.com/subdir/img/picture.png",
        ),
        (
            "https://example.com/subdir/page.html/",
            "/x/X/picture.png?query=123#fragment",
            "https://example.com/x/X/picture.png?query=123#fragment",
        ),
        (
            "https://example.com/subdir/page.html",
            "https:/www.reestrnpa.gov.ua/REESTR/RNAweb.nsf/wpage/doc_card?OpenDocument&ID=689A2315F9D6C1D7C22587CF004A5E58",
            "https://www.reestrnpa.gov.ua/REESTR/RNAweb.nsf/wpage/doc_card?OpenDocument&ID=689A2315F9D6C1D7C22587CF004A5E58",
        ),
        (
            "https://example.com/subdir/page.html",
            "https://confetti.dev",
            "https://confetti.dev",
        ),
        (
            "https://example.com",
            "/test",
            "https://example.com/test",
        ),
        (
            "https://www.eeaa.gov.eg/Laws/58/index",
            R"https://www.eeaa.gov.eg\Uploads/Laws/Files\20221018140838825.pdf",
            "https://www.eeaa.gov.eg/Uploads/Laws/Files/20221018140838825.pdf",
        ),
        (
            # Base HTTP and link HTTPS
            "http://example.com",
            "https://example.com/image.png",
            "https://example.com/image.png",
        ),
        (
            # Base and link HTTP
            "http://example.com",
            "http://example.com/image.png",
            "http://example.com/image.png",
        ),
        (
            # Base HTTPS and link HTTP
            "https://example.com",
            "http://example.com/image.png",
            "http://example.com/image.png",
        ),
        (
            # Normalizing artifacts from not joining urls properly
            "https://evp.nc.gov/",
            "https://evp.nc.gov//solicitations/details/?id=a4382140-25b6-ee11-a569-001dd804ec4b",
            "https://evp.nc.gov/solicitations/details/?id=a4382140-25b6-ee11-a569-001dd804ec4b",
        ),
        (
            "https://www.bidbuy.illinois.gov/bso/view/search/external/advancedSearchBid.xhtml?openBids=true",
            "/bso/external/bidDetail.sdo?docId=24-497DVA-MANTE-B-41187&external=true&parentUrl=close",
            "https://www.bidbuy.illinois.gov/bso/external/bidDetail.sdo?docId=24-497DVA-MANTE-B-41187&external=true&parentUrl=close",
        ),
        (
            "https://www.bidbuy.illinois.gov/bso/view/search/external/advancedSearchBid.xhtml?openBids=true",
            "//example.com",
            "https://example.com",
        ),
        (
            # Escape spaces in URL
            "https://www.emarketplace.state.pa.us",
            "https://www.emarketplace.state.pa.us/FileDownload.aspx?file=6100061149/Solicitation_1.docx&OriginalFileName=Worker Protection Form .docx",
            "https://www.emarketplace.state.pa.us/FileDownload.aspx?file=6100061149/Solicitation_1.docx&OriginalFileName=Worker%20Protection%20Form%20.docx",
        ),
        (
            "file://Users/username/Downloads/harambe_com/index.mhtml",
            "https://harambe.com/detail_page",
            "https://harambe.com/detail_page",
        ),
        (
            "https://primoliquors.com/products/cutwater-espresso-martini-12oz-4-pack-can",
            "//primoliquors.com/cdn/shop/files/oxcuimdzbczobdeo248f.png?v=1727810919&width=1946",
            "https://primoliquors.com/cdn/shop/files/oxcuimdzbczobdeo248f.png?v=1727810919&width=1946",
        ),
        (
            "https://www.example.com",
            "s3://deworkd-local-files/5a9a3ac5-a572-494d-a888-8a065e6e3878/579c900dbfb97dae2fbf329cf6b3411a.pdf;https://example.com/",
            "s3://deworkd-local-files/5a9a3ac5-a572-494d-a888-8a065e6e3878/579c900dbfb97dae2fbf329cf6b3411a.pdf;https://example.com/",
        ),
    ],
)
def test_normalize_url(base_path, url, expected):
    assert normalize_url(url, base_path) == expected


@pytest.mark.parametrize(
    "input_url, expected_url",
    [
        ("http://example.com", "http://example.com"),
        ("https://example.com", "https://example.com"),
        ("http:/example.com", "http://example.com"),
        ("https:/example.com", "https://example.com"),
        ("https/example.com", "https/example.com"),
        ("www.example.com", "www.example.com"),
    ],
)
def test_sanitize_scheme(input_url, expected_url):
    assert sanitize_scheme(input_url) == expected_url



import json

from harambe_core.observer.serialization_observer import SerializationObserver


class JsonStdoutObserver(SerializationObserver):
    def __init__(self):
        super().__init__(sink=print, serializer=json.dumps)



import re
from datetime import datetime
from typing import Any, Union

import dateparser
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

common_non_specific_dates = {
    "tbd",
    "tba",
    "tbc",
    "open until filled",
    "open until contracted",
    "to be confirmed",
    "as soon as possible",
    "subject to change",
    "not specified",
    "ongoing",
    "continuous",
    "rolling basis",
    "pending approval",
    "n/a",
    "indefinite",
    "immediate",
    "unknown",
}


class ParserTypeDate:
    def __new__(cls) -> Any:
        return Annotated[Union[str, None], AfterValidator(cls.validate_type)]

    @staticmethod
    def validate_type(date: str) -> Union[str, None]:
        # Cast to string incase the date is a datetime float/number
        date = str(date)

        # Trim whitespaces
        date = date.strip()

        # Make sure it's not an empty string
        if len(date) == 0:
            raise ValueError("Empty input")

        if date.lower() in common_non_specific_dates:
            return None
        # Attempt to parse date string using dateparser
        parsed_date = dateparser.parse(date)

        if parsed_date is None:
            # Remove timezone abbreviation in parentheses if present
            date = re.sub(r"\s*\(.*\)$", "", date).strip()

            # List of datetime formats to try
            datetime_formats = [
                "%m/%d/%Y %I:%M:%S %p",  # 4/30/2024 09:00:02 AM
                "%Y-%m-%dT%H:%M:%S",  # 2024-04-30T09:00:02
                "%Y-%m-%d %H:%M:%S",  # 2024-04-30 09:00:02
                "%B %d, %Y - %I:%M%p",  # May 14, 2024 - 2:00pm
                "%m/%d/%Y",  # 4/30/2024
            ]

            # Attempt to parse using datetime with specific formats
            for date_format in datetime_formats:
                try:
                    parsed_date = datetime.strptime(date, date_format)
                    break
                except ValueError:
                    continue

            if parsed_date is None:
                raise ValueError(f"Unable to parse input as date: {date}")

        # Return the date in ISO 8601 format
        return parsed_date.isoformat()



import re
from typing import Any, Optional, Union

from price_parser import Price
from pydantic import BeforeValidator
from typing_extensions import Annotated, Literal

from harambe_core.parser.constants import CURRENCY_MAP, PRICE_NOT_AVAILABLE_PHRASES
from harambe_core.parser.type_currency import ParserTypeCurrency

# this will capture only the amount
NUMBER_PATTERN = re.compile(r"-?.?\d[\d,-\.]*")

PriceOutputType = dict[
    Literal["currency", "raw_currency", "amount", "raw_price"],
    Optional[Union[str, float]],
]


class ParserTypePrice:
    def __new__(cls, required: bool = True) -> Any:
        validator = BeforeValidator(cls.validate_price)
        base = PriceOutputType

        return Annotated[
            base if required else base | None,
            validator,
        ]

    @staticmethod
    def validate_price(
        value: Union[str, float, int, None],
    ) -> Optional[PriceOutputType]:
        if isinstance(value, (float, int)):
            return {
                "currency": None,
                "raw_currency": None,
                "amount": float(value),
                "raw_price": str(value),
            }

        if not value or not (value_str := str(value).strip()):
            return None

        if ParserTypePrice._is_price_not_available(value_str):
            return None

        amount = ParserTypePrice._extract_amount(value_str)
        if not isinstance(amount, (float, int)):
            return amount

        # not using price.amount because it does not handle negative amounts yet
        price = Price.fromstring(value_str)
        currency_code = ParserTypePrice._get_currency_code(price.currency)
        return {
            "currency": currency_code,
            "raw_currency": price.currency,
            "amount": amount,
            "raw_price": value_str,
        }

    @staticmethod
    def _is_price_not_available(value: str) -> bool:
        return value.lower() in PRICE_NOT_AVAILABLE_PHRASES

    @staticmethod
    def _get_currency_code(value: str) -> Optional[str]:
        if not value:
            return None

        for key, currency_code in CURRENCY_MAP.items():
            if key.upper() == value.upper():
                return currency_code

        return None

    @staticmethod
    def _extract_amount(value: str) -> float | None | Any:
        match = NUMBER_PATTERN.findall(value)

        if len(match) != 1:
            raise ValueError("Multiple price amounts found")

        return ParserTypeCurrency.validate_currency(match[0])



RESERVED_PREFIX = "harambe_reserved_"

PRICE_NOT_AVAILABLE_PHRASES = [
    "ask for pricing",
    "call for price",
    "check price in store",
    "coming soon",
    "contact for price",
    "contact us for price",
    "currently unavailable",
    "email for price",
    "free",
    "get a quote",
    "inquire for price",
    "login for price",
    "login to see price",
    "n/a",
    "n/a (not available)",
    "no price available",
    "not priced",
    "out of stock",
    "please call",
    "please contact",
    "price coming soon",
    "price hidden",
    "price not available",
    "price not disclosed",
    "price on application",
    "price on request",
    "price tbd (to be determined)",
    "price unavailable",
    "price upon inquiry",
    "price upon request",
    "pricing not provided",
    "quote",
    "request a quote",
    "see details for price",
    "see price after login",
    "sold out",
    "tbd",
    "tdb",
    "temporarily unavailable",
    "unavailable price",
    "variable price",
    "varies",
]

CURRENCY_MAP = {
    "ALL": "ALL",
    "AFN": "AFN",
    "ARS": "ARS",
    "AWG": "AWG",
    "AUD": "AUD",
    "AZN": "AZN",
    "BSD": "BSD",
    "BBD": "BBD",
    "BYN": "BYN",
    "BZD": "BZD",
    "BMD": "BMD",
    "BOB": "BOB",
    "BAM": "BAM",
    "BWP": "BWP",
    "BGN": "BGN",
    "BRL": "BRL",
    "BND": "BND",
    "KHR": "KHR",
    "CAD": "CAD",
    "KYD": "KYD",
    "CLP": "CLP",
    "CNY": "CNY",
    "COP": "COP",
    "CRC": "CRC",
    "CUP": "CUP",
    "CZK": "CZK",
    "DKK": "DKK",
    "DOP": "DOP",
    "XCD": "XCD",
    "EGP": "EGP",
    "SVC": "SVC",
    "EUR": "EUR",
    "FKP": "FKP",
    "FJD": "FJD",
    "GHS": "GHS",
    "GIP": "GIP",
    "GTQ": "GTQ",
    "GGP": "GGP",
    "GYD": "GYD",
    "HNL": "HNL",
    "HKD": "HKD",
    "HUF": "HUF",
    "ISK": "ISK",
    "INR": "INR",
    "IDR": "IDR",
    "IRR": "IRR",
    "IMP": "IMP",
    "ILS": "ILS",
    "JMD": "JMD",
    "JPY": "JPY",
    "JEP": "JEP",
    "KZT": "KZT",
    "KPW": "KPW",
    "KRW": "KRW",
    "KGS": "KGS",
    "LAK": "LAK",
    "LBP": "LBP",
    "LRD": "LRD",
    "MKD": "MKD",
    "MYR": "MYR",
    "MUR": "MUR",
    "MXN": "MXN",
    "MNT": "MNT",
    "MZN": "MZN",
    "NAD": "NAD",
    "NPR": "NPR",
    "ANG": "ANG",
    "NZD": "NZD",
    "NIO": "NIO",
    "NGN": "NGN",
    "NOK": "NOK",
    "OMR": "OMR",
    "PKR": "PKR",
    "PAB": "PAB",
    "PYG": "PYG",
    "PEN": "PEN",
    "PHP": "PHP",
    "PLN": "PLN",
    "QAR": "QAR",
    "RON": "RON",
    "RUB": "RUB",
    "SHP": "SHP",
    "SAR": "SAR",
    "RSD": "RSD",
    "SCR": "SCR",
    "SGD": "SGD",
    "SBD": "SBD",
    "SOS": "SOS",
    "ZAR": "ZAR",
    "LKR": "LKR",
    "SEK": "SEK",
    "CHF": "CHF",
    "SRD": "SRD",
    "SYP": "SYP",
    "TWD": "TWD",
    "THB": "THB",
    "TTD": "TTD",
    "TRY": "TRY",
    "TVD": "TVD",
    "UAH": "UAH",
    "AED": "AED",
    "GBP": "GBP",
    "USD": "USD",
    "UYU": "UYU",
    "UZS": "UZS",
    "VEF": "VEF",
    "VND": "VND",
    "YER": "YER",
    "ZWD": "ZWD",
    "GEL": "GEL",
    "$": "USD",
    "€": "EUR",
    "£": "GBP",
    "¥": "JPY",
    "₹": "INR",
    "₩": "KRW",
    "A$": "AUD",
    "C$": "CAD",
    "NZ$": "NZD",
    "S$": "SGD",
    "HK$": "HKD",
    "R$": "BRL",
    "Z$": "ZWL",
    "₽": "RUB",
    "₺": "TRY",
    "₦": "NGN",
    "₫": "VND",
    "₱": "PHP",
    "₡": "CRC",
    "₭": "LAK",
    "₮": "MNT",
    "₲": "PYG",
    "฿": "THB",
    "₪": "ILS",
    "৳": "BDT",
    "zł": "PLN",
    "៛": "KHR",
    "kr": "NOK",
    "kr.": "SEK",
    "Kč": "CZK",
    "RM": "MYR",
    "R": "ZAR",
    "د.إ": "AED",
    "د.م.": "MAD",
    "د.ك": "KWD",
    "﷼": "IRR",
    "Bs": "VEF",
    "Bs.": "VEF",
    "S/.": "PEN",
    "B/.": "PAB",
    "Lek": "ALL",
    "؋": "AFN",
    "₼": "AZN",
    "Br": "BYN",
    "BZ$": "BZD",
    "$b": "BOB",
    "KM": "BAM",
    "P": "BWP",
    "лв": "BGN",
    "RD$": "DOP",
    "¢": "GHS",
    "J$": "JMD",
    "ден": "MKD",
    "₨": "MUR",
    "Gs": "PYG",
    "lei": "RON",
    "Дин.": "RSD",
    "NT$": "TWD",
    "TT$": "TTD",
    "₴": "UAH",
    "$U": "UYU",
    "DH": "DZD",
    "Ft": "HUF",
    "MK": "MWK",
    "MT": "MZN",
    "L": "RON",
    "T": "KZT",
    "VT": "VUV",
    "Q": "GTQ",
    "ƒ": "ANG",
    "Rp": "IDR",
    "S": "SOS",
    "₾": "GEL",
}



import re
from typing import Any

import phonenumbers
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

phone_number_formats = [
    r"^\d{3,11}$",  # 911 & 11111111111
    r"^\d{3}[\s.-]?\d{4}$",  # 456-7890
    r"^(\(?\d{1,3}\)?[\s.-])?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}$",  # +1 (628) 555-3456 & (+1) 415-155-1555
    r"^\(\+?\d{1,3}\)\s\d{10}(\s\([Ee]xtension:?\s\d{1,4}\))?$",  # (+4) 1111111111 (Extension: 323) & (4) 1111111111 (extension: 323)
    r"^(\(?\d{1,3}\)?\s?)?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}(,?\s*([Ee]xt(?:ension)?\.?|[Ee]xtension:?)\s*:?\s*\d{1,4})?$",  # Covers all other extension formats
]

ordered_text_to_strip = [
    ["tel", "phone", "fax", "mobile", "cell"],
    ["number"],
    [":", " ", "-"],
]


class ParserTypePhoneNumber:
    def __new__(cls) -> Any:
        return Annotated[str, AfterValidator(cls.validate_type)]

    @staticmethod
    def validate_type(number: str) -> str:
        formatted_number = number.strip().lower().replace("/", "-")

        for text_chunk in ordered_text_to_strip:
            for text in text_chunk:
                formatted_number = formatted_number.lstrip(text).strip()

        # First, try using the phonenumbers library
        try:
            phone_number = phonenumbers.parse(
                formatted_number, None
            )  # 'None' implies no specific region

            if phonenumbers.is_valid_number(phone_number):
                # Return the phone number in international format
                return phonenumbers.format_number(
                    phone_number, phonenumbers.PhoneNumberFormat.INTERNATIONAL
                )
        except phonenumbers.phonenumberutil.NumberParseException:
            pass

        # If phonenumbers library fails, fall back to regex validation
        # Remove plus sign
        formatted_number = formatted_number.replace("+", "")
        # Attempt to parse phone number
        for phone_number_format in phone_number_formats:
            regex = re.compile(phone_number_format)
            if regex.match(formatted_number):
                return formatted_number

        raise ValueError(
            f"Unable to parse input as phone number. Original input: {number}. Post processing: {formatted_number}"
        )



from functools import wraps
from typing import Any, Callable

from harambe_core.parser.constants import RESERVED_PREFIX

Func = Callable[..., Any]


class ExpressionEvaluator:
    __builtins__ = {}

    def __init__(self):
        self._functions = {}

    def __contains__(self, function_name: str) -> bool:
        return (
            function_name.upper() in self._functions
            or function_name.upper() in self.__builtins__
        )

    def __getitem__(self, function_name: str) -> Func:
        return (
            self._functions.get(function_name.upper())
            or self.__builtins__[function_name.upper()]
        )

    def __call__(self, func_name: str, *args: Any, **kwargs: Any) -> Any:
        return self[func_name](*args, **kwargs)

    def evaluate(self, expression: str, obj: Any) -> Any:
        expression = expression.strip()

        if not expression:
            raise ValueError("Empty expression")

        if "(" not in expression and ")" not in expression:
            expression = f"NOOP({expression})"

        func_name = expression.split("(", 1)[0].strip().upper()
        if not func_name:
            raise ValueError("Invalid function name")

        if func_name not in self:
            raise ValueError(f"Unknown function: {func_name}")

        remaining = expression[len(func_name) :].strip()
        if not (remaining.startswith("(") and remaining.endswith(")")):
            raise SyntaxError(f"Invalid function call syntax: {expression}")

        args_str = remaining[1:-1]

        # Parse arguments handling nested functions
        args = []
        current_arg = ""
        paren_count = 0

        for char in args_str:
            if char == "(" and current_arg.strip():
                paren_count += 1
            elif char == ")":
                paren_count -= 1
            elif char == "," and paren_count == 0:
                if current_arg.strip():
                    args.append(current_arg.strip())
                current_arg = ""
                continue

            current_arg += char

        if paren_count != 0:
            raise SyntaxError(f"Invalid function call syntax: {expression}")

        if current_arg.strip():
            args.append(current_arg.strip())

        evaluated_args = []
        for arg in args:
            arg = arg.strip()
            if ExpressionEvaluator._is_string_literal(arg):
                evaluated_args.append(arg[1:-1])
            elif "(" in arg:
                evaluated_args.append(self.evaluate(arg, obj))
            else:
                evaluated_args.append(ExpressionEvaluator._get_field_value(arg, obj))

        return self(func_name, *evaluated_args)

    def define_function(self, func_name: str):
        return self._wrap(func_name, self._functions)

    @classmethod
    def define_builtin(cls, func_name: str):
        return cls._wrap(func_name, cls.__builtins__)

    @staticmethod
    def _is_string_literal(arg: str):
        return (arg.startswith("'") and arg.endswith("'")) or (
            arg.startswith('"') and arg.endswith('"')
        )

    @staticmethod
    def _get_field_value(field_path: str, obj: Any) -> Any:
        parts = field_path.strip().split(".")
        current = obj

        for part in parts:
            if part.startswith("model_"):
                part = RESERVED_PREFIX + part

            if "[" in part and "]" in part:
                split = part.split("[")
                part = split[0]

                if part not in current or not isinstance(current[part], list):
                    return None

                idx = split[1].rstrip("]")
                current = current[part][int(idx)]
            elif attr := getattr(current, part, None):
                current = attr
            elif isinstance(current, dict):
                current = current.get(part)
            else:
                return None

        return current

    @staticmethod
    def _wrap(
        func_name: str, function_store: dict[str, Callable[..., Any]]
    ) -> Callable[..., Any]:
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except AttributeError:
                    return None

            function_store[func_name.upper()] = wrapper
            return wrapper

        return decorator
